/***************************************************************************\
  "error.h" - Error Handling Functions (Global Phase)
  Revision 9; 10 Nov 2002 - 21 Dec 2002
  Copyright (C) 2002  Kevin P. Albrecht
\***************************************************************************/

#ifndef ERROR_H
#define ERROR_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //

// "Program" Error Codes (Controller) //
#define EP_INVALID_ARG               "P01: Invalid command line argument"
#define EP_INVALID_ARGS              "P02: Invalid command line arguments"
#define EP_MISSING_ARG               "P03: Missing command line argument"
#define EP_FILE_OPEN_FAILED          "P04: Error opening file"

// Error Codes for "Front-End" (Phases A, B, and C) //
#define EF_UNDEFINED_TOKEN           "F01: Undefined token"
#define EF_UNDECLARED_IDENTIFIER     "F02: Undeclared identifier"
#define EF_UNEXPECTED_TOKEN          "F03: Unexpected token"
#define EF_UNEXPECTED_EOF            "F04: Unexpected end-of-file"
#define EF_UNNECESSARY_TYPE_CONV     "F05: Unnecessary type conversion"
#define EF_MISSING_EOL               "F06: Missing end-of-line"
#define EF_MISSING_ASSIGNMENT_OP     "F07: Missing assignment operator"
#define EF_MISSING_RIGHT_PAREN       "F08: Missing right parenthesis"
#define EF_INVALID_EXPRESSION        "F09: Invalid expression"
#define EF_ALREADY_DEFINED           "F10: Identifier already defined"
#define EF_TYPE_INCOMPATIBLE         "F11: Type is incompatible"
#define EF_INVALID_PARSE_POINT       "F12: Invalid parsing point"
#define EF_INCOMPLETE_COMMENT        "F13: Comment is never closed"
#define EF_CONST_ASSIGNMENT          "F14: Cannot assign a new value to a constant"
#define EF_MAX_SCOPES_REACHED        "F15: Maximum concurrent scopes reached"
#define EF_MIN_SCOPES_REACHED        "F16a: Minimum scopes reached; Cannot leave global scope"
#define EF_INDENT_ERROR              "F16b: Decreased indent must match indent of outer scope"
#define EF_MISSING_TYPE_DEF_OP       "F17: Missing type definition operator"
#define EF_INVALID_WHITESPACE        "F18: Invalid whitespace character"
#define EF_NOT_VALID_TYPE            "F19: Not a valid type"
#define EF_MISSING_EXPRESSION        "F20: Missing expression"

// Error Codes for "Intermediate Code Manager" (Phase D) //
#define EI_WRITE_AFTER_RESET         "I01: Intermediate code write attempt after reset"
#define EI_READ_BEFORE_RESET         "I02: Intermediate code read attempt before reset"
#define EI_GOTO_BEFORE_RESET         "I03: Intermediate code goto attempt before reset"
#define EI_CURR_INDEX_OUT_OF_BOUNDS  "I04: Current intermediate code index out of bounds"
#define EI_MAX_TEMPS_EXCEEDED        "I05: Maximum number of temporary variables exceeded"
#define EI_MIN_TEMPS_REACHED         "I06: Minimum number of temporary variables reached"
#define EI_MAX_UNIQUE_NAMES_EXCEEDED "I07: Maximum number of unique names exceeded"

// Error Codes for "Runtime Engine" (Phase E) //
#define ER_INVALID_INSTRUCTION       "R01: Invalid instruction encountered"
#define ER_JUMP_NOT_DEFINED          "R02: Jump was not defined by a label"
#define ER_DIV_BY_ZERO               "R03: Divide by zero attempted"
#define ER_POP_MISMATCHED_TYPE       "R04: Destination of pop must match type on stack"
#define ER_POP_STACK_EMPTY           "R05: Attempted to pop from empty stack"

/**** Not yet implemented ****/
//#define EF_NESTING_TOO_DEEP          "F12: Nesting too deep"
//#define EF_REPEAT_DECLARATIVE        "F14: Only one declarative sequence allowed per program"
//#define EF_SUB_ALREADY_DECLARED      "F15: Subroutine already declared"
//#define EF_VARIABLE_IN_DECLARATION   "F16: Cannot use variable in its own declaration"
//#define ER_RUNTIME_ERROR             "R02: Runtime error"
//#define ER_DIVIDE_BY_ZERO            "R03: Divide by zero"
//#define ER_STACK_OVERFLOW            "R04: Runtime stack overflow error"
//#define ER_STACK_EMPTY               "R05: Runtime stack empty"
//#define ER_INVALID_EXEC_POINT        "R06: Invalid execution point"
//#define ER_EXCESSIVE_GOTOS           "R07: Too many gotos were executed"
//#define ER_INVALID_ASSIGNMENT        "R08: Invalid assignment"
//#define ER_DESTINATION_NOT_DEFINED   "R09: Attempted to goto undefined destination"
//#define ER_VARIABLE_ALREADY_DEFINED  "R10: Variable previously defined"
//#define ER_VARIABLE_NOT_DEFINED      "R11: Variable not defined"
//#define ER_FILE_ALREADY_OPEN         "R12a: File already open"
//#define ER_FILE_NOT_OPEN             "R12b: File not open"
//#define ER_ATTEMPT_INPUT_ON_OUTFILE  "R13: Cannot perform input on an output file"
//#define ER_ATTEMPT_OUTPUT_ON_INFILE  "R14: Cannot perform output on an input file"
//#define ES_RAT_OUT_OF_RANGE          "Rational number out of range"
//#define ES_INT_OUT_OF_RANGE          "Integer out of range"
//#define ES_TOO_MANY_DIGITS           "Too many digits"
//#define ES_NOT_IMPLEMENTED           "Feature not yet implemented"

// Effect Codes //
#define EFF_IGNORE 0
#define EFF_RECOVER 1
#define EFF_TERMINATE 2

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdio.h>
#include <stdlib.h>

// Local //
#include "global.h"
#include "k_string.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

void err (bool lines, char *sError, char *sClarification);
void errm (char *s1, char *s2);
void erre (int nEffect);
void verbose (bool lines, char *s1, char *s2, char *s3, char *s4);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// err
//// Display an error
void err (bool lines, char *sError, char *sClarification)
{
  gnErrors++;

  fprintf (stderr, "** Error ");

  if (lines) fprintf (stderr, "(%s:%d) ", gsCurrentFilename, gnCurrentLine);
  fprintf (stderr, "%s", sError);
  if (sClarification) fprintf (stderr, ": %s", sClarification);
  fprintf (stderr, "\n");
}

//// errm
//// Display an additional error message
void errm (char *s1, char *s2)
{
  if (s1) fprintf (stderr, "** (continued...) %s", s1);
  if (s2) fprintf (stderr, ": %s", s2);
  fprintf (stderr, "\n");
}

//// erre
//// Display and/or execute an error effect
void erre (int nEffect)
{
  switch (nEffect)
  {
    case (EFF_IGNORE):
      break;
    case (EFF_RECOVER):
      break;
    case (EFF_TERMINATE):
      fprintf (stderr, "** Fatal Error: Terminating program...\n");
      exit (0);
      break;
    default:
      fprintf (stderr, "** Invalid Error Effect: %d\n", nEffect);
      break;
  }
}

//// verbose
//// Print too much info :-)
void verbose (bool lines, char *s1, char *s2, char *s3, char *s4)
{
  if (!gbVerbose) return;

  fprintf (stderr, "<<");
  if (lines) fprintf (stderr, "#%d:", gnCurrentLine);
  if (s1) if (strlen(s1) > 0) fprintf (stderr, "%s", s1);
  if (s2) if (strlen(s2) > 0) fprintf (stderr, ":%s", s2);
  if (s3) if (strlen(s3) > 0) fprintf (stderr, ":%s", s3);
  if (s4) if (strlen(s4) > 0) fprintf (stderr, ":%s", s4);
  fprintf (stderr, ">>\n");
}

#endif
