/***************************************************************************\
  "icm.h" - Intermediate Code Manager (Phase D)
  Revision 7; 10 Nov 2002 - 13 Dec 2002
  Copyright (C) 2002  Kevin P. Albrecht
\***************************************************************************/

#ifndef ICM_H
#define ICM_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define ICM_MAX_STATEMENTS 1024

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdlib.h>

// Local //
#include "global.h"
#include "error.h"
#include "table.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //

typedef enum // Three-Address Code Operations
{
  tacNOP,
  tacLINE, // marks a statement holding a linenumber
  tacJUMP, tacSTOP, tacCOND,
  tacXIR, tacXIS, tacXRI, tacXRS, tacXSI, tacXSR,
  tacSET, tacPUSH, tacPOP,
  tacADD, tacSUB, tacMULT, tacDIV, tacMOD, tacPOW,
  tacNEG,
  tacEQ, tacNEQ, tacLT, tacGT, tacLTEQ, tacGTEQ,
  tacOR, tacAND,
  tacCAT, tacSEQ, tacSNEQ,
  tacCIN, tacCOUT
} TAC;

char *gvsTACStrings[] =
{
  "nop",
  "line",
  "jump", "stop", "cond",
  "xir", "xis", "xri", "xrs", "xsi", "xsr",
  "set", "push", "pop",
  "add", "sub", "mult", "div", "mod", "pow",
  "neg",
  "eq", "neq", "lt", "gt", "lteq", "gteq",
  "or", "and",
  "cat", "seq", "sneq",
  "cin", "cout"
};

typedef struct
{
  TAC tac;
  Type type;
  Symbol *a;
  Symbol *b;
  Symbol *c;
  int number; // misc. number holder    tac         number
              //                        ===         ======
              //                        tacLINE     line number
} Statement;

Statement *statement_new (TAC aTac, Symbol *aA, Symbol *aB, Symbol *aC);
void statement_del (Statement *statement);

Statement** gvProgram; // vector of statements
int gxLast;            // index of last statement in the gvProgram vector
int gxCurr;            // index of current statement in the gvProgram vector

void icm_init ();
void icm_deinit ();
int icm_put (Statement *statement);
void icm_putline ();
void icm_reset ();
void icm_goto (int x);
Statement *icm_getnext ();
Statement *icm_get (int x);

// Internal //
bool mReset;

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// statement_new
Statement *statement_new (TAC aTac, Symbol *aA, Symbol *aB, Symbol *aC)
{
  Statement *temp;

  temp = (Statement *) malloc (sizeof (Statement));
  temp->tac = aTac;
  temp->a = aA;
  temp->b = aB;
  temp->c = aC;
  temp->number = 0;

  return temp;
}

//// statement_del
void statement_del (Statement *statement)
{
  if (statement) free (statement);
}

//// icm_init
void icm_init ()
{
  int i;

  gvProgram = (Statement **) malloc (sizeof (Statement *) * ICM_MAX_STATEMENTS);
  for (i = 0; i < ICM_MAX_STATEMENTS; i++)
  {
    gvProgram[i] = NULL;
  }

  gxLast = 0;
  gxCurr = 0;
  mReset = false;
}

//// icm_deinit
void icm_deinit ()
{
  int i;

  for (i = 0; i < ICM_MAX_STATEMENTS; i++)
  {
    if (gvProgram[i]) statement_del (gvProgram[i]);
  }
  free (gvProgram);
}

//// icm_put
int icm_put (Statement *statement)
{
  if (mReset == false)
  {
    if ((gxCurr + 1) > ICM_MAX_STATEMENTS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    gvProgram[gxCurr] = statement;
    gxCurr++;
    gxLast++;
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    erre (EFF_TERMINATE);
  }

  return (gxCurr - 1);
}

//// icm_putline
void icm_putline ()
{
  Statement *statement;
  Symbol *symbol;

  if (mReset == false)
  {
    if ((gxCurr + 1) > ICM_MAX_STATEMENTS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    symbol = sym_new (gsCurrentFilename);
    sym_set (symbol, repLiteral, tStr, false, gsCurrentFilename);
    statement = statement_new (tacLINE, symbol, NULL, NULL);
    statement->number = gnCurrentLine;
    gvProgram[gxCurr] = statement;
    gxCurr++;
    gxLast++;
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    erre (EFF_TERMINATE);
  }
}

//// icm_reset
void icm_reset ()
{
  gxCurr = 0;
  mReset = true;
}

//// icm_goto
void icm_goto (int x)
{
  if (mReset == true)
  {
    if ((x > gxLast) || (x > ICM_MAX_STATEMENTS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    gxCurr = x;
  }
  else
  {
    err (true, EI_GOTO_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
  }
}

//// icm_getnext
Statement *icm_getnext ()
{
  if (mReset == true)
  {
    if ((gxCurr + 1) > gxLast)
    {
      return NULL; // there are no more statements, so return NULL
    }

    return gvProgram[gxCurr++];
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
    return NULL;
  }
}

//// icm_get
Statement *icm_get (int x)
{
  if (mReset == true)
  {
    if ((x > gxLast) || (x > ICM_MAX_STATEMENTS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    return gvProgram[x];
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
    return NULL;
  }
}

#endif
