/***************************************************************************\
  "re.h" - Runtime Engine (Phase E)
  Revision 9; 10 Nov 2002 - 15 Dec 2002
  Copyright (C) 2002  Kevin P. Albrecht
\***************************************************************************/

#ifndef RE_H
#define RE_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define RE_INPUT_LINE_LEN 256

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdio.h>
#include <math.h>

// Local //
#include "global.h"
#include "error.h"
#include "icm.h"
#include "k_string.h"
#include "rstack.h"
#include "global_conv.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// Internal //
Statement *mSM;
RStack *mRS;
TAC mTAC;

// External //
void re_init ();
void re_deinit ();
void re_run ();

// Internal - Run //
void mRunStatement ();
void mRunJump ();
void mRunConditional ();
void mRunConversion ();
void mRunAssignment ();
void mRunStackOperation ();
void mRunUnaryArith ();
void mRunBinaryArith ();
void mRunBinaryString ();
void mRunConsoleInput ();
void mRunConsoleOutput ();

// Internal - Special //
void mGetStatement ();

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// re_init
//// Initialize re module
void re_init ()
{
  mRS = rstack_new ();
}

//// re_deinit
//// Deinitialize re module
void re_deinit ()
{
  rstack_del (mRS);
}

//// re_run
void re_run ()
{
  mGetStatement();
  gnCurrentLine = 0;

  while (mTAC != tacSTOP)
  {
    mRunStatement ();
    mGetStatement ();
  }
}

/////////
// Run //
/////////

//// mRunStatement
void mRunStatement ()
{
  switch (mTAC)
  {
    case (tacNOP):
      // do nothing! //
      break;
    case (tacLINE):
      gnCurrentLine = mSM->number;
      gsCurrentFilename = mSM->a->value;
      break;
    case (tacJUMP):
      mRunJump ();
      break;
    case (tacCOND):
      mRunConditional ();
      break;
    case (tacXIR):
    case (tacXIS):
    case (tacXRI):
    case (tacXRS):
    case (tacXSI):
    case (tacXSR):
      mRunConversion ();
      break;
    case (tacSET):
      mRunAssignment ();
      break;
    case (tacPUSH):
    case (tacPOP):
      mRunStackOperation ();
      break;
    case (tacNEG):
      mRunUnaryArith ();
      break;
    case (tacADD):
    case (tacSUB):
    case (tacMULT):
    case (tacDIV):
    case (tacMOD):
    case (tacPOW):
    case (tacEQ):
    case (tacNEQ):
    case (tacLT):
    case (tacGT):
    case (tacLTEQ):
    case (tacGTEQ):
    case (tacOR):
    case (tacAND):
      mRunBinaryArith ();
      break;
    case (tacCAT):
    case (tacSEQ):
    case (tacSNEQ):
      mRunBinaryString ();
      break;
    case (tacCIN):
      mRunConsoleInput ();
      break;      
    case (tacCOUT):
      mRunConsoleOutput ();
      break;
    default:
      err (true, ER_INVALID_INSTRUCTION, gvsTACStrings[mTAC]);
      erre (EFF_TERMINATE);
      break;
  }
}

//// mRunJump
void mRunJump ()
{
  // FORMAT: jump (null) (null) (null) (index number) //

  if (mSM->a->constant == true)
  {
    icm_goto (*(int*)(mSM->a->value));
  }
  else
  {
    err (true, ER_JUMP_NOT_DEFINED, mSM->a->key);
    erre (EFF_TERMINATE);
  }
}

//// mRunConditional
void mRunConditional ()
{
  // FORMAT: cond (boolean) (true destination) (false destination) 0 //

  if ( (*(int*)(mSM->a->value)) != 0 )
  {
    icm_goto (*(int*)(mSM->b->value));
  }
  else
  {
    icm_goto (*(int*)(mSM->c->value));
  }
}

//// mRunConversion
void mRunConversion ()
{
  // FORMAT: x__ (dest) (source) (null) 0 //

  int ntemp;
  float ftemp;
  char *stemp;

  ntemp = 0;
  ftemp = 0.0;
  stemp = NULL;

  free (mSM->a->value); // free the current value

  switch (mTAC)
  {
    case (tacXIR):
      ftemp = (float) *(int*)mSM->b->value;
      mSM->a->value = k_new_f (ftemp);
      break;
    case (tacXIS):
      stemp = conv_is ( *(int*)mSM->b->value );
      mSM->a->value = k_clone (stemp);
      break;
    case (tacXRI):
      ntemp = (int) *(float*)mSM->b->value;
      mSM->a->value = k_new_n (ntemp);
      break;
    case (tacXRS):
      stemp = conv_fs ( *(float*)mSM->b->value );
      mSM->a->value = k_clone (stemp);
      break;
    case (tacXSI):
      ntemp = conv_si ( (char*) mSM->b->value );
      mSM->a->value = k_new_n (ntemp);
      break;
    case (tacXSR):
      ftemp = conv_sf ( (char*) mSM->b->value );
      mSM->a->value = k_new_f (ftemp);
      break;
  }
  
  //BUGFIX: for some reason, this causes error
  //if (stemp) free (stemp);
}

//// mRunAssignment
void mRunAssignment ()
{
  // FORMAT: set (dest) (source) (null) 0 //

  int ntemp;
  float ftemp;

  ntemp = 0;
  ftemp = 0.0;

  free (mSM->a->value); // free the current value

  // Put a copy of B's value in A //
  switch (mSM->a->type)
  {
    case (tInt):
      if (mSM->b->type == tRat) ntemp = (int) *(float*)mSM->b->value;
      else                      ntemp = *(int*)mSM->b->value;
      mSM->a->value = k_new_n (ntemp);
      break;
    case (tRat):
      if (mSM->b->type == tInt) ftemp = (float) *(int*)mSM->b->value;
      else                      ftemp = *(float*)mSM->b->value;
      mSM->a->value = k_new_f (ftemp);
      break;
    case (tStr):
      mSM->a->value = k_clone (mSM->b->value);
      break;
  }
}

//// mRunStackOperation
void mRunStackOperation ()
{
  // PUSH //
  if (mTAC == tacPUSH)
  {
    rstack_push (mRS, mSM->a);
  }

  // POP //
  else if (mTAC == tacPOP)
  {
    // types are the same //
    if (mSM->a->type == rstack_peektype(mRS))
    {
      free (mSM->a->value);

      mSM->a->value = rstack_pop (mRS);

      if (mSM->a->value == NULL)
      {
        err (true, ER_POP_STACK_EMPTY, NULL);
        erre (EFF_TERMINATE);
      }
    }
    // types are different //
    else
    {
      err (true, ER_POP_MISMATCHED_TYPE, gvsTypeStrings[rstack_peektype(mRS)]);
      erre (EFF_TERMINATE);
    }
  }
}

//// mRunUnaryArith
void mRunUnaryArith ()
{
  // FORMAT: neg (dest) (source) (null) 0 //

  int ntemp;
  float ftemp;

  ntemp = 0;
  ftemp = 0.0;

  free (mSM->a->value); // free the current value

  // Perform the operation //
  switch (mSM->a->type)
  {
    case (tInt):
      ntemp = *(int*)mSM->b->value;
      mSM->a->value = k_new_n (-ntemp);
      break;
    case (tRat):
      ftemp = *(float*)mSM->b->value;
      mSM->a->value = k_new_f (-ftemp);
      break;
  }
}

//// mRunBinaryArith
void mRunBinaryArith ()
{
  // FORMAT: ___ (dest) (source_b) (source_c) 0 //

  float fb, fc, fresult;

  // Put the float value of B in 'fb' //
  if (mSM->b->type != tRat)
  {
    fb = (float) *(int*)mSM->b->value;
  }
  else
  {
    fb = *(float*)mSM->b->value;
  }

  // Put the float value of C in 'fc' //
  if (mSM->c->type != tRat)
  {
    fc = (float) *(int*)mSM->c->value;
  }
  else
  {
    fc = *(float*)mSM->c->value;
  }

  switch (mTAC)
  {
    case (tacADD):
      fresult = fb + fc;
      break;
    case (tacSUB):
      fresult = fb - fc;
      break;
    case (tacMULT):
      fresult = fb * fc;
      break;
    case (tacDIV):
      if (fc == 0.0) {
        err (true, ER_DIV_BY_ZERO, NULL);
        erre (EFF_TERMINATE);
      } else  {
        fresult = fb / fc;
      }
      break;
    case (tacMOD):
      if (fc == 0.0) {
        err (true, ER_DIV_BY_ZERO, NULL);
        erre (EFF_TERMINATE);
      } else  {
        fresult = (float) ((int)(fb) % (int)(fc));
      }
      break;
    case (tacPOW):
      fresult = (float) pow (fb, fc);
      break;
    case (tacEQ):
      fresult = (float) (fb == fc);
      break;
    case (tacNEQ):
      fresult = (float) (fb != fc);
      break;
    case (tacLT):
      fresult = (float) (fb < fc);
      break;
    case (tacGT):
      fresult = (float) (fb > fc);
      break;
    case (tacLTEQ):
      fresult = (float) (fb <= fc);
      break;
    case (tacGTEQ):
      fresult = (float) (fb >= fc);
      break;
    case (tacOR):
      fresult = (float) (fb || fc);
      break;
    case (tacAND):
      fresult = (float) (fb && fc);
      break;
  }

  free (mSM->a->value); // free the current value

  // Put the value of 'fresult' into A //
  switch (mSM->a->type)
  {
    case (tInt):
      mSM->a->value = k_new_n ( (int) fresult );
      break;
    case (tRat):
      mSM->a->value = k_new_f ( fresult );
      break;
  }
}

//// mRunBinaryString
void mRunBinaryString ()
{
  char *temp;

  // FORMAT: cat (dest) (source_b) (source_c) 0 //
  if (mTAC == tacCAT)
  {
    temp = k_new (strlen (mSM->b->value) + strlen(mSM->c->value) + 1);
    k_copy (temp, "");
    k_cat (temp, mSM->b->value);
    k_cat (temp, mSM->c->value);

    k_del (mSM->a->value);
    mSM->a->value = temp;
  }
  else if (mTAC == tacSEQ)
  {
    if (strcmp ((char*)(mSM->b->value), (char*)(mSM->c->value)) == 0)
      mSM->a->value = k_new_f (1.0);
    else
      mSM->a->value = k_new_f (0.0);
  }
  else if (mTAC == tacSNEQ)
  {
    if (strcmp ((char*)(mSM->b->value), (char*)(mSM->c->value)) != 0)
      mSM->a->value = k_new_f (1.0);
    else
      mSM->a->value = k_new_f (0.0);
  }
}

//// mRunConsoleInput
void mRunConsoleInput ()
{
  char input[RE_INPUT_LINE_LEN];

  free (mSM->a->value); // free the current value

  gets (input);

  switch (mSM->a->type)
  {
    case (tInt):
      mSM->a->value = k_new_n ( conv_si (input) );
      break;
    case (tRat):
      mSM->a->value = k_new_f ( conv_sf (input) );
      break;
    case (tStr):
      mSM->a->value = k_clone (input);
      break;
  }
}

//// mRunConsoleOutput
void mRunConsoleOutput ()
{
  switch (mSM->a->type)
  {
    case (tInt):
      printf ("%d", *(int *)mSM->a->value);
      break;
    case (tRat):
      printf ("%f", *(float *)mSM->a->value);
      break;
    case (tStr):
      printf ("%s", (char *)mSM->a->value);
      break;
  }
}

//////////////
// Special  //
//////////////

void mGetStatement ()
{
  mSM = icm_getnext();
  mTAC = mSM->tac;
}

#endif

